#!/usr/bin/perl -w

package Source::Repository::Mass;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use File::Path qw(make_path remove_tree);
use RPM qw/evrcmp vercmp/;

use Source::Repository::RPM;
use Source::Repository::Matcher;

require Exporter;

our $VERSION=0.06;

our @ISA = qw(Exporter);
our @EXPORT = qw/$DESTPATH $ORIGINPATH @ORIGINSKIP @DESTSKIP/;
our ($DESTPATH, $ORIGINPATH);

our @ORIGINSKIP=();
our @DESTSKIP=();
my %ORIGINSKIP;
my %DESTSKIP;

our $matcher_class='Source::Repository::Matcher';
our $origin_class='Source::Repository::RPM';
our $dest_class='Source::Repository::RPM';
our @ORIGIN_DEFAULT_OPTS;
our @DEST_DEFAULT_OPTS;

our $verbose=0;
our $strategy = 'all';

my ($opt_transaction, $opt_taskfile, $opt_all, $opt_destmode, $help);
my ($opt_origin_namelist, $opt_dest_namelist, $opt_namelist);
my ($opt_origin_repolist, $opt_dest_repolist);
my ($opt_origin_pathlist, $opt_dest_pathlist);
my ($opt_origin_cachedir, $opt_dest_cachedir);
# TODO
my ($opt_origin_skiplist, $opt_dest_skiplist);

our ($arg_timestampdir,$opt_skip_on_error);
our $arg_timestamp_expire=7; # in days


our %STRATEGYMAP=(
    # quick synonyms
    'force' => \&origin2dest,
    'f' => \&origin2dest,
    'a' => \&origin2dest,
    'm' => \&originskipdest,
    'o' => \&destskiporigin,
    'u' => \&destup,
    'n' => \&newer,

    'all' => \&origin2dest,
    'missing' => \&originskipdest,
    'orphaned' => \&destskiporigin,
    'destup' => \&destup,
    'newer' => \&newer,
    'older' => \&older,
    'both' => \&both,
    # S(ilent) names
    'allS' => \&origin2destS,
    'missingS' => \&originskipdestS,
    'orphanedS' => \&destskiporiginS,
    'destupS' => \&destupS,
    'newerS' => \&newerS,
    'olderS' => \&olderS,
    'bothS' => \&bothS,

    # old names
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destskiporigin' => \&destskiporigin,

    # verup, relup
    'verup' => \&verup,
    'relup' => \&relup,

    # print help
    'help' => \&help_strategy,
    );

our %STRATEGYDESCR=(
    'all' => 'any package',
    'missing' => 'packages in origin, missing in dest', 
    'orphaned' => 'packages in dest, missing in origin',
    'destup' => 'packages in dest, older than origin (i.e. need update)',
    'newer' => 'packages in origin, new and newer than dest',
    'older' => 'packages in origin, older than dest',
    'both' => 'packages in origin, present in dest',
    'relup' => 'destup, limited release update',
    'verup' => 'destup, limited to version update',

    );

my @LONGOPT = (
	"all"	=> \$opt_all,
	"dest=s"	=> \$DESTPATH,
	"dest-cachedir=s"	=> \$opt_dest_cachedir,
	"dest-namelist=s"	=> \$opt_dest_namelist,
	"dest-repolist=s"	=> \$opt_dest_repolist,
	"dest-path=s"		=> \$opt_dest_pathlist,
	"dest-skiplist=s"	=> \$opt_dest_skiplist,
	"input-dest"	=> sub {$opt_destmode=1},
	"input-origin"	=> sub {$opt_destmode=0},
	"origin=s"	=> \$ORIGINPATH,
	"origin-cachedir=s"	=> \$opt_origin_cachedir,
	"origin-namelist=s"	=> \$opt_origin_namelist,
	"origin-repolist=s"	=> \$opt_origin_repolist,
	"origin-path=s"		=> \$opt_origin_pathlist,
	"origin-skiplist=s"	=> \$opt_origin_skiplist,
	"skip-on-error"	=> \$opt_skip_on_error,
	"strategy=s"	=> \$strategy,
	"timestampdir=s"	=> \$arg_timestampdir,
	"timestamp-expire=f"	=> \$arg_timestamp_expire,
	"transaction"	=> \$opt_transaction,
	"tasklist=s"	=> \$opt_taskfile,
	"namelist=s"	=> \$opt_namelist,

	"help"		=> \$help,
	"v|verbose+"	=> \$verbose,
	'q|quiet'   => sub { $verbose = 0 },
    );

our @EXT_OPTS;
our @EXT_OPTS_USAGE;
our @EXT_LONGOPT_CLASSES;

sub new {
    my $class = shift;
    my $self = {
	'-originopts' => [],
	'-destopts' => [],
	'-cliargs' => [],
	'-packages' => [],
	@_
    };
    bless $self, $class;
    my $result = GetOptions (
	@LONGOPT,
	@EXT_OPTS,
	@{$self->{'-cliargs'}},
	# should be always the last (under the map)
	map {$_->get_longopt} @EXT_LONGOPT_CLASSES);

    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }

    $self->{VERBOSE}=$verbose;

    my (@defrepo_opts, @origin_opts, @dest_opts);
    push @defrepo_opts, -namelist => $opt_namelist if $opt_namelist;
    push @origin_opts, -namelist => $opt_origin_namelist if $opt_origin_namelist;
    push @dest_opts, -namelist => $opt_dest_namelist if $opt_dest_namelist;
    push @origin_opts, -repolist => $opt_origin_repolist if $opt_origin_repolist;
    push @dest_opts, -repolist => $opt_dest_repolist if $opt_dest_repolist;
    push @origin_opts, -cachedir => $opt_origin_cachedir if $opt_origin_cachedir;
    push @dest_opts, -cachedir => $opt_dest_cachedir if $opt_dest_cachedir;
    push @origin_opts, -pathlist => $opt_origin_pathlist if $opt_origin_pathlist;
    push @dest_opts, -pathlist => $opt_dest_pathlist if $opt_dest_pathlist;

    $self->{-originrepo}=$origin_class->new(
	-default_opt=>$ORIGINPATH,
	@ORIGIN_DEFAULT_OPTS,
	$opt_destmode ? () : @defrepo_opts,
	@origin_opts,
	@{$self->{-originopts}},
	);
    $self->{-destrepo}=$dest_class->new(
	-default_opt=>$DESTPATH,
	@DEST_DEFAULT_OPTS,
	$opt_destmode ? @defrepo_opts : (),
	@dest_opts,
	@{$self->{-destopts}},
	);

    %ORIGINSKIP=map {$_ => 1} @ORIGINSKIP;
    %DESTSKIP=map {$_ => 1} @DESTSKIP;

    my $def_repo = $opt_destmode ? $self->{-destrepo} : $self->{-originrepo};
    if ($opt_taskfile) {
	open (my $fn, '<', $opt_taskfile) or die "can't open file $opt_taskfile\n";
	while (my $line=<$fn>) {
	    my @list = map{$def_repo->expand_glob($_)} &Source::Repository::__line2array($line);
	    next if scalar @list == 0;
	    if (scalar @list == 1) {
		push @{$self->{'-packages'}}, @list;
	    } else {
		# transaction
		push @{$self->{'-packages'}}, [@list];
	    }
	}
	close($fn);
    }

    make_path($arg_timestampdir) if defined $arg_timestampdir;

    push @{$self->{'-packages'}}, $opt_transaction ? [@ARGV] : @ARGV;
    push @{$self->{'-packages'}}, $def_repo->names() if $opt_all;
    return $self;
}

sub usage {
    print "usage: $0 [ options ] [ list of origin rpms/names ]

Example:
upgrade Sisyphus java packages using jpackage
originmass --dest /var/ftp/pub/Sisyphus/files/SRPMS --origin /var/ftp/pub/jpackage/6.0/generic/free/SRPMS

General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet

Mass processing options:
	--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
	--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
	--skip-on-error	ignore individual package failures
	--strategy 'strategy'	of mass operations (default is $strategy).
";&_print_strategies();

    foreach my $usage(@EXT_OPTS_USAGE) {
	if (ref $usage ne 'CODE') {
	    die "Oops! not a code: $usage ".(ref $usage eq 'SCALAR' ? $$usage : '');
	} else {
	    print "\n";
	    &$usage;
	}
    }
}

sub _print_strategies {
    print "	Available strategies: 
";
    my %main_name=map {$STRATEGYMAP{$_}=>$_} keys (%STRATEGYDESCR);
    my %synonym=map {$_=>[]} keys (%STRATEGYDESCR);
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	my $mname=$main_name{$STRATEGYMAP{$sname}};
	push@{$synonym{$mname}}, $sname if $mname and not defined $description;
    }
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	if (defined $description) {
	    print "\t\t$sname\t- $description\n";
	    my @synonyms = @{$synonym{$sname}};
	    print "\t\t\t  Synonyms: ".join(',',@synonyms)."\n" if @synonyms;
	}
    }
}

sub originrepo {
    my ($self)=@_;
    return $self->{-originrepo};
}

sub destrepo {
    my ($self)=@_;
    return $self->{-destrepo};
}

sub process {
    my ($self)=@_;
    foreach my $nameobj (@{$self->{'-packages'}}) {
	if (ref $nameobj) {
	    $self->_process_transaction($nameobj);
	} else {
	    $self->_process_name($nameobj);
	}
    }
}

sub _process_name {
    my ($self, $name)=@_;
    my $pkgpair=$self->_get_matcher()->new_pair($name);
    return if $self->__skip_due_to_timestamp($pkgpair);
    return unless $self->is_strategy_allow($pkgpair);
    $self->__mk_timestamp($pkgpair);
    $self->_process_pair_wrapper($pkgpair);
}

sub _process_transaction {
    my ($self, $transaction_ptr)=@_;
    my $transaction_name=$transaction_ptr->[0];
    my $is_in_transaction=0;
    my $transaction_count=0;
    foreach my $name (@$transaction_ptr) {
	my $pkgpair=$self->_get_matcher()->new_pair($name);
	next if $self->__skip_due_to_timestamp($pkgpair);
	next unless $self->is_strategy_allow($pkgpair);
	if (not $is_in_transaction) {
	    $self->init_transaction($transaction_name);
	    $is_in_transaction=1;
	}
	$self->__mk_timestamp($pkgpair);
	$self->_process_pair_wrapper($pkgpair, $transaction_name, ++$transaction_count);
    }
    $self->done_transaction($transaction_name,$transaction_count) if $is_in_transaction;
}

sub _get_matcher {
    my $self=shift;
    my $matcher_obj=$self->{-matcher};
    unless ($matcher_obj) {
	$matcher_obj = $matcher_class->new(
	-originrepo => $self->{-originrepo}, 
	-destrepo => $self->{-destrepo}, 
	-destmode=>$opt_destmode,
	);
	$self->{-matcher}=$matcher_obj;
    }
    return $matcher_obj;
}

sub init_transaction {
    my ($self,$transaction_name)=@_;
}

sub done_transaction {
    my ($self,$transaction_name)=@_;
}

sub _process_pair_wrapper {
    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
    eval {
	$self->process_pair($pkgpair,$transaction_name,$transaction_count);
    };
    if ($@) {
	die 'FAILED: '.$pkgpair->originname.": $@

You can ignore this error with --skip-on-error option\n" unless $opt_skip_on_error;
	print STDERR 'FAILED: '.$pkgpair->originname.": $@";
    };
}

###############################################################
# The method process_pair should be implemented in subclasses;
# disable the prototype implementation completely to allow 
# multiple inheritance tricks with Transform::{Embedded,PassThrough}
###############################################################
#sub process_pair {
#    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
#    print STDERR "process_pair: $self\n";
#    ...
#}

sub is_strategy_allow {
    my ($self,$pkgpair)=@_;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    unless ($strategy_allow) {
	&_print_strategies();
	die "strategy $strategy not found.\n" ;
    }
    my $originname=$pkgpair->originname;
    my $destname=$pkgpair->destname;
    return 0 if $ORIGINSKIP{$originname} or $DESTSKIP{$destname};
    return 1 if &{$strategy_allow}($self, $pkgpair->origin, $pkgpair->dest, $originname, $destname);
    return 0;
}

sub cmp_origin_dest {
    my ($self, $origin, $dest)=@_;
    return 0 unless $origin;
    return 1 unless $dest;
    return vercmp($origin->version, $origin->release, $dest->version, $dest->release)
}

sub __skip_due_to_timestamp {
    my ($self,$pkgpair)=@_;
    return 0 unless defined $arg_timestampdir;
    my $timestamp_file=&__srpmfile2timestamp($pkgpair->origin->file);
    return 1 if -M $timestamp_file and ($arg_timestamp_expire==0 or -M $timestamp_file < $arg_timestamp_expire);
    return 0;
}

sub __mk_timestamp {
    my ($self,$pkgpair)=@_;
    return unless defined $arg_timestampdir;
    &__run('touch',&__srpmfile2timestamp($pkgpair->origin->file));
}

#------------------------------------------

#------------------------------------------
#    private strategies
#------------------------------------------

sub help_strategy {
    &_print_strategies();
    exit(1);
}

sub origin2dest {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1;
}

sub origin2destS {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1;
}

sub destup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub destupS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub newer {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub newerS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub both {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}


sub bothS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}

sub originskipdest {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "found dest RPM $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub originskipdestS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "found dest RPM $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub older {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)<0;
    return 0;
}

sub olderS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)<0;
    return 0;
}

sub destskiporigin {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: dest RPM $destname not found.\n" unless $dest;
    return 0 unless $dest;
    print STDERR "found origin RPM $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

sub destskiporiginS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: dest RPM $destname not found.\n" unless $dest;
    return 0 unless $dest;
    #print STDERR "found origin RPM $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

## ########
# release up; version is the same, something other (release) is different
## ########
sub relup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version eq $origin->version and $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

## ########
# version up; versions are different
## ########
sub verup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version ne $origin->version and $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}



#------------------------------------------

#------------------------------------------
#    private utilities
#------------------------------------------

sub __srpmfile2timestamp {
    my ($rpm)=@_;
    return $arg_timestampdir.'/'.basename($rpm).'.stamp';
}

sub __run {
    print STDERR "running: ", join(' ',@_),"\n" if $Source::Repository::Mass::verbose;
    system(@_)==0 or Carp::confess "command failed: ".join(' ',@_);
}


1;
